# Optimization engine interface
# Should be free of SimPy.Simulation objects

import os
from lpsolve55 import *
from scipy import *

##DEBUG, only used when not called from commutesim.py
##global numnodes
##numnodes = 3


basedir = 'output'
subdir = 0
dstdir = basedir + '/run' + str(subdir)

def initsched(initdemand, nn, nt=5):
    """ Write a file for the optimizer to load the scenario """
    global subdir, basedir, dstdir
    subdir += 1
    dstdir = basedir + '/run' + str(subdir)    
    if not os.path.exists(dstdir): os.mkdir(dstdir)
    f = open(dstdir + '/initsched.txt', 'w')
    f.write('# Scenario initialization automatically generated by schedopt.py\n')
    #nn = 5   # number of nodes in transit network
    ns = 2  # number of vehicle types / sizes
    #nt = 5  # number of timesteps
    f.write('nn = ' + str(nn) + '\n')  # Number of nodes in transit system
    f.write('ns = ' + str(ns) + '\n' )    # Number of aircraft types / sizes
    f.write('nt = ' + str(nt) + '\n') # Number of transfer timesteps used
    f.write('maxaircraft = 40\n')  # Maximum number of vehicles in operation
    f.write('desirednumac = 20\n') # preferred number of vehicles in operation
    f.write('cyclicschedule = 0\n') # Enforce ending positions == starting positions
    f.write('optcriteria = max: numsrv;\n')  # goal statement
##    f.write('optcriteria = max: profit;\n')  # goal statement
    f.write('dir = ' + dstdir + '\n')


    for i in range(0,nn):
        f.write('ncap_' + str(i) + ' = 6\n')
        f.write('sc_' + str(i) + ' = 10\n')
        for k in range(0,nn):
            if (i != k):
                for t in range(0,nt):
                    if (initdemand[t][i][k] != 0):  # Save some filespace
                        f.write('initdemand_' +str(t) +'_' + str(i) + '_' + str(k) + ' = ' + str(initdemand[t][i][k]) + '\n' )
    
    for s in range(0,ns):
        f.write('fc_' + str(s) + ' = ' + str(4 + 10*s) + '\n')
        f.write('aac_' + str(s) + ' = ' + str(10 + 40*s) + '\n')
        f.write('fcap_' + str(s) + ' = ' + str(4 + 20*s) + '\n')
    
    f.close()


def optimize():
    """Initialize the problem setup
    run the schedop.pl script to generate the model
    and finally run the model and return a table with the results"""
    
    global subdir, basedir, dstdir
    
    # Run schedopt.pl script to generate our model
    os.system('schedopt/schedopt.pl ' + dstdir + '/initsched.txt')
    
    # Load the model
    lp = lpsolve('read_LP', dstdir + '/sched.lpm')
    
    # Set a timeout on finding an optimal solution (returning a suboptimal one when breaking prematurely, or a timeout if no feasible solutions have been found)
    lpsolve('set_timeout', lp, 20.0)   # timeout in seconds of computation
    
    # Solve the model
    lpsolve('solve', lp)
    
    # Store results into a dictionary
    colnames = lpsolve('get_col_names', lp)
    vars = lpsolve('get_variables', lp)[0]

    # Store results for debugging purposes
    f = open(dstdir + '/result.txt', 'w')
    for n in colnames:
        f.write( n  + " = " + str(vars[colnames.index(n)]) + '\n')
    f.close()

    return [colnames, vars]    
    

def writeCSV(sd, nn, ts, finalts):
    """export the schedule solution data to a formatted spreadsheet for easy viewing / debugging"""
    global subdir, basedir, dstdir
    f = open ( dstdir + '/state' + str(ts) + '.csv', 'w')
    f.write('Scheduled state\n')
    f.write('Timestep,' + str(ts) +  ' out of ' + str(finalts) + '\n')
    f.write('Summary:\n')
    for k in ('numsrv', 'numac', 'tpr', 'tfc', 'profit', 'spareacp', 'spareacm', 'numacdev'):
        f.write(k + ',' + str(sd[k]) + '\n')
        
    
    for s in range(1,2+1):      ##FIXME: replace with number of vehicle types
        f.write('Vehicle movements, ' + str(s) + '\n')
        f.write(' ,from/to:')
        for j in range(1,nn+1):
            f.write(', ' + str(j))
        f.write('\n')
        for i in range(1,nn+1):
            f.write(' ,' + str(i))
            for j in range(1,nn+1):
                if (i != j):
                    if (ts != finalts):
                        f.write(', ' + str(sd['f_' + str(ts) +'_'+ str(s) +'_'+ str(i) +'_'+ str(j)]))
                    else:
                        f.write(', ')
                else:
                    f.write(', ' + str(sd['ap_' + str(ts) +'_'+ str(s) +'_' + str(i)]))
            f.write('\n')
                 
        f.write('\n')
        
    if (ts != finalts):
        f.write('People movements\n')
        f.write(' ,from/to:')
        for j in range(1,nn+1):
            f.write(', ' + str(j))
        f.write('\n')
        for i in range(1,nn+1):
            f.write(' ,' + str(i))
            for j in range(1,nn+1):
                if (i != j):  ## FIXME: sum over k
                    p=0
                    for k in range(1,nn+1):
                        if (i != k):
                            p+= sd['p_' + str(ts) +'_'+ str(i) +'_'+ str(j) +'_'+ str(k)]
                    f.write(', ' + str(p))
                else:
                    f.write(', ')
            f.write('\n')
        f.write('\n')

    f.write('People pools\n')
    f.write(' ,from/to:')
    for j in range(1,nn+1):
        f.write(', ' + str(j))
    f.write('\n')
    if (ts != 0):
        for i in range(1,nn+1):
            f.write(' ,' + str(i))
            for k in range(1,nn+1):
                if (ts != finalts or (ts == finalts and i == k)):
                    f.write(', ' + str(sd['pp_' + str(ts) +'_'+ str(i) +'_'+ str(k)]))
                else:
                    f.write(', ')
            f.write('\n')
        
##    for i in range(1,nn+1):
##        f.write('\nNode:, ' + str(i) + '\n')
##        
##        for s in range(1,2+1):      ##FIXME: replace with number of vehicle types
##            f.write(' ,Vehicle type ' + str(s) + ', ' + str(sd['ap_' + str(ts) + '_' + str(s) + '_' + str(i)]) + '\n')
##        
##        f.write(' ,Destination:')
##        for k in range(1,nn+1):
##            f.write(', ' + str(k))
##        f.write('\n')
##        
##        for s in range(1,2+1):   ##FIXME: replace with number of vehicle types
##            if (ts != 0):
##                f.write(' ,Vehicle ' + str(s))
##                for j in range(1,nn+1):
##                    if (j != i):
##                        f.write(', ' + str(sd['f_' + str(ts) +'_'+ str(s) +'_'+ str(i) +'_'+ str(j)]))
##                    else:
##                        f.write(', X')
##            f.write('\n')
##        
##        f.write(' ,Passengers leaving')
##        for j in range(1,nn+1):
##            if (ts != 0):
##                if (i != j):
##                    p = 0
##                    for k in range(1,nn+1):
##                        if (j != k and i != k): 
##                            p += sd['p_' + str(ts) +'_'+ str(i) +'_'+ str(j) +'_'+ str(k)]
##                    f.write(', ' + str(p))
##                else:
##                    f.write(', X')
##        f.write('\n')
##        
##        f.write(' ,Passenger pool')
##        for k in range(1,nn+1):
##            if (ts != 0):
##                f.write(', ' + str(sd['pp_' + str(ts) + '_' + str(i) + '_' + str(k)]))
##        f.write('\n')
##        
##        f.write(' ,New passengers') 
##        for k in range(1,nn+1):
##            if (i != k):
##                f.write(', ' + str(sd['ppi_' + str(ts) + '_' + str(i) + '_' + str(k)]))
##            else:
##                f.write(', X')
##        f.write('\n')
        
    f.close()
    
    
class transitSchedule:
    """Abstract representation of transit requests as an intermediary for parsing into the schedule optimizer"""
    def __init__(self,numnodes,timestep=15):
        self.nn = numnodes
        self.ts = timestep    # each schedule optimizer timestep corresponds to 15 minutes of simtime
        # Initialize demand counter between nodes
        self.demand = []  # list of tuples in format: (source,dest,departure time, number of passengers)
        self.dirty = 0  # flag to indicate presence of unprocessed incoming requests
        self.numbins = 0    # number of timesteps used in the last solution
    def addRequest(self, src, dst, deptime, numpassengers=1):
        """Adds a request for passengers travelling from src to dst node with a departure simtime"""
        self.dirty = 1
        self.demand.append((src, dst, deptime, numpassengers))
    def exportDemand(self, curtime):
        """Translate self.demand in simtime to use discrete, relative timesteps for export to schedule optimizer"""
        #Loop through self.demand and preen out reasonable range of deptimes
        times = []  # Create list of departure simtimes
        ##print str(self.demand)  ##DEBUG
        for req in self.demand:
            times.append(req[2])
        # For now, accept the entire range of requests
        mintime = min(times)
        maxtime = max(times)
        duration = maxtime - mintime
        # Convert from curTime to relative discrete timebins
        numbins = duration / self.ts + 4   # integer division, plus extra timesteps to complete transfers
        self.numbins = numbins      # this info is handy for postprocessing
        # Build bins of deptimes, and insert entries for relocations
        ##FIXME: remove leading zero timesteps somehow
        discretedemand = zeros((numbins,self.nn,self.nn))
        for t in range(numbins):
            for reqnum in range(len(times)):    
                if  ((times[reqnum] - mintime) >= self.ts * t) and ((times[reqnum] - mintime) < self.ts * (t+1)):
                    # discretedemand[timestep][srcnode][dstnode] += number of travelers
                    discretedemand[t][self.demand[reqnum][0]][self.demand[reqnum][1]] += self.demand[reqnum][3]
                    #FIXME: find a way to pop that reqnum from self.demand without screwing up the indexing for subsequent passes
        initsched(discretedemand, nn=self.nn, nt=numbins)
        ##print "Numbins : " + str(numbins)    ##DEBUG
        ##print "Discrete Demand : \n" + str(discretedemand)    ##DEBUG
        soln = optimize() 
        ##print "Solution : " + str(soln)     ##DEBUG
        # Cleanup steps
        self.dirty = 0  # reset incoming request flag
        self.demand = []  # FIXME:  clear out only the ones that were addressed
        # Transpose soln from two arrays to one list of key:value pairs
        sa = [] 
        for i in range(len(soln[0])):
            sa.append((soln[0][i], soln[1][i]))
        # Convert solution to a dictionary
        sd = dict(sa)
        return sd
    def printSoln(self,soln):
        """Output the state at each timestep"""
        for t in range(self.numbins+1):
            writeCSV(soln, self.nn, t, self.numbins+1)

    
def testOpt():
    """Test optimizer interface"""    
    nn=numnodes
    nt = 5
    initdemand = zeros((nt,nn,nn))
    
    for i in range(nn):
        for k in range(nn):
            if (i != k):
                for t in range(2):
                    initdemand[t][i][k] = 100

    initsched(initdemand, nn=nn, nt=nt)

    soln = optimize()
    
    # Transpose soln from two arrays to one list of key:value pairs
    sa = [] 
    for i in range(len(soln[0])):
        sa.append((soln[0][i], soln[1][i]))
    # Convert solution to a dictionary
    sd = dict(sa)
    print sd  ##DEBUG

    
def testReq():
      """Test schedule request interface"""
      sched = transitSchedule(numnodes)
      
#testOpt()
#testReq()
